const LOG = require("npm-niosz-log");
const log = LOG.newLog("oracle");
const json = require("npm-niosz-json");
const oracledb = require("oracledb");
if (process.env.LD_LIBRARY_PATH) oracledb.initOracleClient({libDir: process.env.LD_LIBRARY_PATH});
const defaultConfigSize = 1000;
const defaultFetchSize = 500;
const uuid = require("uuid");
const fs = require("fs");
// ****************************************************************************
const checkSelectConfig = function(obj) {
    var resp = {};
    if (typeof obj === 'object' && obj != null) {
        for (prop in obj) 
            resp[prop] = obj[prop];
        if (!resp.configSize) resp.configSize = defaultConfigSize;
        if (!resp.fetchSize) resp.fetchSize = defaultFetchSize;
        if (resp.fetchSize > resp.configSize) resp.fetchSize = resp.configSize;
    }
    return resp;
};
// ****************************************************************************
const generateConnectionConfig = function(obj) {
    var resp = {
        user : null,
        password : null,
        connectString : null
    };
    if (typeof obj === 'object' && obj != null) {
        if (obj.user) resp.user = obj.user;
        if (obj.password) resp.password = obj.password;
        if (obj.connectString) resp.connectString = obj.connectString;
    }
    return resp;
};
// ****************************************************************************
const generateExcuteConfig = function(obj) {
    var resp = {
        outFormat:oracledb.OUT_FORMAT_OBJECT,
        prefetchRows:defaultConfigSize,
        fetchArraySize:defaultConfigSize,
        resultSet:true
    };
    if (typeof obj === 'object' && obj != null) {
        if (obj.configSize) {
            resp.prefetchRows = obj.configSize;
            resp.fetchArraySize = obj.configSize;
        }
    }
    return resp;
};
// ****************************************************************************
const generateFileName = function(i) {
    return i.toString().padStart(10,"0")+".json";
};
// ****************************************************************************
const oraCamelCase = function(column) {
    column = column.toLowerCase();
    column = column.split("_");
    for(var i=1;i<column.length;i++) {
        if (column[i].length>1) {
            column[i] = column[i][0].toUpperCase() + column[i].substring(1);
        } else {
            column[i] = column[i].toUpperCase();
        }
    }
    return column.join("");
};
// ****************************************************************************
const createResultSet = function(folder,fetchs,rows,error) {
    return {
        folder          : folder,
        fetchs          : fetchs,
        remainingFetchs : fetchs,
        rows            : rows,
        remainingRows   : rows,
        error           : error,
        close           : function() { try{fs.rmSync(this.folder,{recursive:true,force:true});return true;}catch{return false;}; },
        fetch           : function(camelCaseOrMapper) {
            if (this.error) return false;
            try {
                var data = json.parseFromFile(this.folder+"/"+generateFileName(this.fetchs-this.remainingFetchs),true);
                this.remainingFetchs--;
                this.remainingRows-=data.length;
                if (camelCaseOrMapper && data.length>0) {
                    if (typeof camelCaseOrMapper == "function") {
                        data = data.map(camelCaseOrMapper);
                    } else {
                        data = data.map((row) => {
                            var camelRow = {};
                            for (prop in row) {
                                const camelProp = oraCamelCase(prop);
                                camelRow[camelProp] = row[prop];
                            }
                            return camelRow;
                        });
                    }
                }
                return data;
            } catch {
                return false;
            }
        }
    };
};
// ****************************************************************************
const oraConnect = async function(connectionConfig) {
    try {
        var oracle = await oracledb.getConnection(connectionConfig);
        log.i("Connesso correttamente al database oracle");
        return oracle;
    } catch(e) {
        log.e("Connessione in errore : " + e);
        return null;
    }
};
// ****************************************************************************
const oraDisconnect = async function(connection,resultSet) {
    try {
        if (resultSet) try { await resultSet.resultSet.close(); } catch {};
        await connection.release();
        log.i("Disconnesso correttamente dal database Oracle");
    } catch(e) {
        log.w("Disconnesso con segnalazioni dal database Oracle:\n"+e);
    }
};
// ****************************************************************************
const oraSelect = async function(selectConfig,queryConfig) {

    // *** check select config
    const selectionConfig = checkSelectConfig(selectConfig);

    // *** generate connection config
    const connectionConfig = generateConnectionConfig(selectionConfig);

    // *** connection to oracle
    var oracle = await oraConnect(connectionConfig);
    if (oracle == null) return null;

    // *** generate execute config
    const executeConfig = generateExcuteConfig(selectionConfig);

    // *** esecuzione query
    var queryData = null;
    try {
        queryData = await oracle.execute(queryConfig,[],executeConfig);
        log.i("Query eseguita correttamente");
    } catch(e) {
        log.e("Errore nell'esecuzione della query : " + e);
        return null;
    }

    // *** prepare folder
    var folderData = null;
    try {
        folderData = "oradata/" + uuid.v4();
        fs.mkdirSync(folderData,{recursive:true});
        log.i("Creata cartella dati : " + folderData);
    } catch(e) {
        log.e("Errore nella creazione cartella dati : " + e);
        return null;
    }

    // *** fetching data
    var fetch = 0;
    var len = 0;
    var eof = false;
    while (!eof) {
        var rows = null;
        try {
            rows = await queryData.resultSet.getRows(selectionConfig.configSize);
        } catch(e) {
            log.e("Errore durante il fetching dei dati : " + e);
            await oraDisconnect(oracle,queryData);
            return createResultSet(folderData,fetch,len,true);
        }
        eof = (!rows.length) ? true : false;
        if (eof) continue;
        try {
            if (selectionConfig.configSize == selectionConfig.fetchSize) {
                json.stringifyToFile(folderData+"/"+generateFileName(fetch),rows);
                fetch++;
                len += rows.length;
            } else {
                var block = [];
                while (rows.length > 0) {
                    block.push(rows.shift());
                    if (block.length == selectionConfig.fetchSize) {
                        json.stringifyToFile(folderData+"/"+generateFileName(fetch),block);
                        fetch++;
                        len += block.length;
                        block = [];
                    }
                }
                if (block.length > 0) {
                    json.stringifyToFile(folderData+"/"+generateFileName(fetch),block);
                    fetch++;
                    len += block.length;
                    block = [];
                }
            }
        } catch(e) {
            log.e("Errore durante la scrittura fetching dati : " + e);
            await oraDisconnect(oracle,queryData);
            return createResultSet(folderData,fetch,len,true);
        }
        log.i("Record processati : " + len);
    }
    log.i("Record totali processati : " + len);

    // *** disconnection from oracle
    await oraDisconnect(oracle,queryData);
    
    // *** create management resultset response
    return createResultSet(folderData,fetch,len,false);

};
// ****************************************************************************
const ora = {
    select : oraSelect
};
// ****************************************************************************
module.exports = ora;